"""Various mergeplan tests
"""

from fediblockhole.blocklist_parser import parse_blocklist
from fediblockhole import merge_blocklists, merge_comments, apply_mergeplan

from fediblockhole.const import SeverityLevel, DomainBlock

datafile01 = "data-suspends-01.csv"
datafile02 = "data-silences-01.csv"
datafile03 = "data-noop-01.csv"

import_fields = [
    'domain',
    'severity',
    'public_comment',
    'private_comment',
    'reject_media',
    'reject_reports',
    'obfuscate'
]

def load_test_blocklist_data(datafiles):

    blocklists = {}

    for df in datafiles:
        with open(df) as fp:
            data = fp.read()
            bl = parse_blocklist(data, 'csv', import_fields)
            blocklists[df] = bl
    
    return blocklists

def test_mergeplan_max():
    """Test 'max' mergeplan"""
    blocklists = load_test_blocklist_data([datafile01, datafile02])

    bl = merge_blocklists(blocklists, 'max')
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.SUSPEND

def test_mergeplan_min():
    """Test 'max' mergeplan"""
    blocklists = load_test_blocklist_data([datafile01, datafile02])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.SILENCE

def test_mergeplan_default():
    """Default mergeplan is max, so see if it's chosen"""
    blocklists = load_test_blocklist_data([datafile01, datafile02])

    bl = merge_blocklists(blocklists)
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.SUSPEND

def test_mergeplan_3_max():
    """3 datafiles and mergeplan of 'max'"""
    blocklists = load_test_blocklist_data([datafile01, datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'max')
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.SUSPEND
        assert bl[key].reject_media == True
        assert bl[key].reject_reports == True
        assert bl[key].obfuscate == True

def test_mergeplan_3_min():
    """3 datafiles and mergeplan of 'min'"""
    blocklists = load_test_blocklist_data([datafile01, datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.NONE
        assert bl[key].reject_media == False
        assert bl[key].reject_reports == False
        assert bl[key].obfuscate == False

def test_mergeplan_noop_v_silence_max():
    """Mergeplan of max should choose silence over noop"""
    blocklists = load_test_blocklist_data([datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'max')
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.SILENCE

def test_mergeplan_noop_v_silence_min():
    """Mergeplan of min should choose noop over silence"""
    blocklists = load_test_blocklist_data([datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    for key in bl:
        assert bl[key].severity.level == SeverityLevel.NONE

def test_merge_public_comment():
    blocklists = load_test_blocklist_data([datafile01, datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    assert bl['public-comment.example.org'].public_comment == 'This is a public comment'

def test_merge_private_comment():
    blocklists = load_test_blocklist_data([datafile01, datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    assert bl['private-comment.example.org'].private_comment == 'This is a private comment'

def test_merge_public_comments():
    blocklists = load_test_blocklist_data([datafile01, datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    assert bl['diff-comment.example.org'].public_comment == 'Suspend public comment, Silence public comment, Noop public comment'

def test_merge_duplicate_comments():
    """The same comment on multiple sources shouldn't get added
    """
    blocklists = load_test_blocklist_data([datafile01, datafile02, datafile03])

    bl = merge_blocklists(blocklists, 'min')
    assert len(bl) == 13

    # Nope, this breaks. Need to rethink duplicate comment merge.
    # assert bl['2diff-comment.example.org'].public_comment == 'Suspend comment 1, Public duplicate'

def test_merge_comments_none():

    a = None
    b = None

    r = merge_comments(a, b)

    assert r == ''

def test_merge_comments_empty():

    a = ''
    b = ''

    r = merge_comments(a, b)

    assert r == ''

def test_merge_comments_left():

    a = 'comment to merge'
    b = ''

    r = merge_comments(a, b)

    assert r == 'comment to merge'

def test_merge_comments_right():

    a = ''
    b = 'comment to merge'

    r = merge_comments(a, b)

    assert r == 'comment to merge'

def test_merge_comments_same():

    a = 'comment to merge'
    b = 'comment to merge'

    r = merge_comments(a, b)

    assert r == 'comment to merge'

def test_merge_comments_diff():

    a = 'comment A'
    b = 'comment B'

    r = merge_comments(a, b)

    assert r == 'comment A, comment B'

def test_merge_comments_dups():

    a = "boring, nazis, lack of moderation, flagged, special"
    b = "spoon, nazis, flagged, lack of moderation, happy, fork"

    r = merge_comments(a, b)

    assert r == 'boring, nazis, lack of moderation, flagged, special, spoon, happy, fork'

def test_mergeplan_same_min_bools_false():
    """Test merging with mergeplan 'max' and False values doesn't change them
    """
    a = DomainBlock('example.org', 'noop', '', '', False, False, False)
    b = DomainBlock('example.org', 'noop', '', '', False, False, False)

    r = apply_mergeplan(a, b, 'max')

    assert r.reject_media == False
    assert r.reject_reports == False
    assert r.obfuscate == False

def test_mergeplan_same_min_bools_true():
    """Test merging with mergeplan 'max' and True values doesn't change them
    """
    a = DomainBlock('example.org', 'noop', '', '', True, False, True)
    b = DomainBlock('example.org', 'noop', '', '', True, False, True)

    r = apply_mergeplan(a, b, 'max')

    assert r.reject_media == True
    assert r.reject_reports == False
    assert r.obfuscate == True

def test_mergeplan_max_bools():
    a = DomainBlock('example.org', 'suspend', '', '', True, True, True)
    b = DomainBlock('example.org', 'noop', '', '', False, False, False)

    r = apply_mergeplan(a, b, 'max')

    assert r.reject_media == True
    assert r.reject_reports == True
    assert r.obfuscate == True


"""Tests of the CSV parsing
"""

from fediblockhole.blocklist_parser import BlocklistParserJSON, parse_blocklist
from fediblockhole.const import DomainBlock, BlockSeverity, SeverityLevel

datafile = 'data-mastodon.json'

def load_data():
    with open(datafile) as fp:
        return fp.read()

def test_json_parser():

    data = load_data()
    parser = BlocklistParserJSON()
    bl = parser.parse_blocklist(data)

    assert len(bl) == 10
    assert bl[0].domain == 'example.org'
    assert bl[1].domain == 'example2.org'
    assert bl[2].domain == 'example3.org'
    assert bl[3].domain == 'example4.org'

    assert bl[0].severity.level == SeverityLevel.SUSPEND
    assert bl[1].severity.level == SeverityLevel.SILENCE
    assert bl[2].severity.level == SeverityLevel.SUSPEND
    assert bl[3].severity.level == SeverityLevel.NONE

def test_ignore_comments():

    data = load_data()
    parser = BlocklistParserJSON()
    bl = parser.parse_blocklist(data)

    assert len(bl) == 10
    assert bl[0].domain == 'example.org'
    assert bl[1].domain == 'example2.org'
    assert bl[2].domain == 'example3.org'
    assert bl[3].domain == 'example4.org'

    assert bl[0].public_comment == ''
    assert bl[0].private_comment == ''

    assert bl[2].public_comment == ''
    assert bl[2].private_comment == ''


"""Parse various blocklist data formats
"""
from typing import Iterable
from .const import DomainBlock, BlockSeverity

import csv
import json

import logging
log = logging.getLogger('fediblockhole')

class BlocklistParser(object):
    """
    Base class for parsing blocklists
    """
    preparse = False

    def __init__(self, import_fields: list=['domain', 'severity'], 
        max_severity: str='suspend'):
        """Create a Parser

        @param import_fields: an optional list of fields to limit the parser to.
            Ignore any fields in a block item that aren't in import_fields.
        """
        self.import_fields = import_fields
        self.max_severity = BlockSeverity(max_severity)

    def preparse(self, blockdata) -> Iterable:
        """Some raw datatypes need to be converted into an iterable
        """
        raise NotImplementedError

    def parse_blocklist(self, blockdata) -> dict[DomainBlock]:
        """Parse an iterable of blocklist items
        @param blocklist: An Iterable of blocklist items
        @returns: A dict of DomainBlocks, keyed by domain
        """
        if self.preparse:
            blockdata = self.preparse(blockdata)

        parsed_list = []
        for blockitem in blockdata:
            parsed_list.append(self.parse_item(blockitem))
        return parsed_list
    
    def parse_item(self, blockitem) -> DomainBlock:
        """Parse an individual block item

        @param blockitem: an individual block to be parsed
        @param import_fields: fields of a block we will import
        """
        raise NotImplementedError

class BlocklistParserJSON(BlocklistParser):
    """Parse a JSON formatted blocklist"""
    preparse = True

    def preparse(self, blockdata) -> Iterable:
        """Parse the blockdata as JSON
        """
        return json.loads(blockdata)

    def parse_item(self, blockitem: str) -> DomainBlock:
        # Remove fields we don't want to import
        origitem = blockitem.copy()
        for key in origitem:
            if key not in self.import_fields:
                del blockitem[key]

        # Convert dict to NamedTuple with the double-star operator
        # See: https://docs.python.org/3/tutorial/controlflow.html#tut-unpacking-arguments
        block = DomainBlock(**blockitem)
        if block.severity > self.max_severity:
            block.severity = self.max_severity
        return block

class BlocklistParserCSV(BlocklistParser):
    """ Parse CSV formatted blocklists

    The parser expects the CSV data to include a header with the field names.
    """
    preparse = True

    def preparse(self, blockdata) -> Iterable:
        """Use a csv.DictReader to create an iterable from the blockdata
        """
        return csv.DictReader(blockdata.split('\n'))

    def parse_item(self, blockitem: dict) -> DomainBlock:
        # Coerce booleans from string to Python bool
        # FIXME: Is this still necessary with the DomainBlock object?
        for boolkey in ['reject_media', 'reject_reports', 'obfuscate']:
            if boolkey in blockitem:
                blockitem[boolkey] = str2bool(blockitem[boolkey])

        # Remove fields we don't want to import
        origitem = blockitem.copy()
        for key in origitem:
            if key not in self.import_fields:
                log.debug(f"ignoring field '{key}'")
                del blockitem[key]

        # Convert dict to DomainBlock with the double-star operator
        # See: https://docs.python.org/3/tutorial/controlflow.html#tut-unpacking-arguments
        block = DomainBlock(**blockitem)
        if block.severity > self.max_severity:
            block.severity = self.max_severity
        return block

class RapidBlockParserCSV(BlocklistParserCSV):
    """ Parse RapidBlock CSV blocklists

    RapidBlock CSV blocklists are just a newline separated list of domains.
    """
    def preparse(self, blockdata) -> Iterable:
        """Prepend a 'domain' field header to the data
        """
        log.debug(f"blockdata: {blockdata[:100]}")
        blockdata = ''.join(["domain\r\n", blockdata])

        return csv.DictReader(blockdata.split('\r\n'))

class RapidBlockParserJSON(BlocklistParserJSON):
    """Parse RapidBlock JSON formatted blocklists
    """
    def preparse(self, blockdata) -> Iterable:
        rb_dict = json.loads(blockdata)
        # We want to iterate over all the dictionary items
        return rb_dict['blocks'].items()

    def parse_item(self, blockitem: tuple) -> DomainBlock:
        """Parse an individual item in a RapidBlock list
        """
        # Each item is a tuple of:
        # (domain, {dictionary of attributes})
        domain = blockitem[0]

        # RapidBlock has a binary block level which we map
        # to 'suspend' if True, and 'noop' if False.
        isblocked = blockitem[1]['isBlocked']
        if isblocked:
            severity = 'suspend'
        else:
            severity = 'noop'
        
        if 'public_comment' in self.import_fields:
            public_comment = blockitem[1]['reason']
        else:
            public_comment = ''
        
        # There's a 'tags' field as well, but we can't
        # do much with that in Mastodon yet

        block = DomainBlock(domain, severity, public_comment)
        if block.severity > self.max_severity:
            block.severity = self.max_severity

        return block

def str2bool(boolstring: str) -> bool:
    """Helper function to convert boolean strings to actual Python bools
    """
    boolstring = boolstring.lower()
    if boolstring in ['true', 't', '1', 'y', 'yes']:
        return True
    elif boolstring in ['', 'false', 'f', '0', 'n', 'no']:
        return False
    else:
        raise ValueError(f"Cannot parse value '{boolstring}' as boolean")

FORMAT_PARSERS = {
    'csv': BlocklistParserCSV,
    'json': BlocklistParserJSON,
    'rapidblock.csv': RapidBlockParserCSV,
    'rapidblock.json': RapidBlockParserJSON,
}

# helper function to select the appropriate Parser
def parse_blocklist(
    blockdata,
    format="csv",
    import_fields: list=['domain', 'severity'],
    max_severity: str='suspend'):
    """Parse a blocklist in the given format
    """
    parser = FORMAT_PARSERS[format](import_fields, max_severity)
    log.debug(f"parsing {format} blocklist with import_fields: {import_fields}...")
    return parser.parse_blocklist(blockdata)


"""Tests of the Rapidblock CSV parsing
"""

from fediblockhole.blocklist_parser import RapidBlockParserCSV, parse_blocklist
from fediblockhole.const import DomainBlock, BlockSeverity, SeverityLevel

csvdata = """example.org\r\nsubdomain.example.org\r\nanotherdomain.org\r\ndomain4.org\r\n"""
parser = RapidBlockParserCSV()

def test_basic_rapidblock():

    bl = parser.parse_blocklist(csvdata)
    assert len(bl) == 4
    assert bl[0].domain == 'example.org'
    assert bl[1].domain == 'subdomain.example.org'
    assert bl[2].domain == 'anotherdomain.org'
    assert bl[3].domain == 'domain4.org'

def test_severity_is_suspend():
    bl = parser.parse_blocklist(csvdata)

    for block in bl:
        assert block.severity.level == SeverityLevel.SUSPEND


""" Utility functions for tests
"""
from fediblockhole import setup_argparse, augment_args

def shim_argparse(testargv: list=[], tomldata: str=None):
    """Helper function to parse test args
    """
    ap = setup_argparse()
    args = ap.parse_args(testargv)
    args = augment_args(args, tomldata)
    return args


""" Constant objects used by FediBlockHole
"""
import enum
from typing import NamedTuple, Optional, TypedDict
from dataclasses import dataclass

import logging
log = logging.getLogger('fediblockhole')

class SeverityLevel(enum.IntEnum):
    """How severe should a block be? Higher is more severe.
    """
    NONE = enum.auto()
    SILENCE = enum.auto()
    SUSPEND = enum.auto()

class BlockSeverity(object):
    """A representation of a block severity

    We add some helpful functions rather than using a bare IntEnum
    """

    def __init__(self, severity:str=None):
        self._level = self.str2level(severity)

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, value):
        if isinstance(value, SeverityLevel):
            self._level = value
        elif type(value) == type(''):
            self._level = self.str2level(value)
        else:
            raise ValueError(f"Invalid level value '{value}'")

    def str2level(self, severity:str=None):
        """Convert a string severity level to an internal enum"""

        if severity in [None, '', 'noop']:
            return SeverityLevel.NONE

        elif severity in ['silence']:
            return SeverityLevel.SILENCE
        
        elif severity in ['suspend']:
            return SeverityLevel.SUSPEND

        else:
            raise ValueError(f"Invalid severity value '{severity}'")

    def __repr__(self):
        return f"'{str(self)}'"

    def __str__(self):
        """A string version of the severity level
        """
        levelmap = {
            SeverityLevel.NONE: 'noop',
            SeverityLevel.SILENCE: 'silence',
            SeverityLevel.SUSPEND: 'suspend',
        }
        return levelmap[self.level]

    def __lt__(self, other):
        if self._level < other._level:
            return True

    def __gt__(self, other):
        if self._level > other._level:
            return True

    def __eq__(self, other):
        if other is not None and self._level == other._level:
            return True

    def __le__(self, other):
        if self._level <= other._level:
            return True

    def __ge__(self, other):
        if self._level >= other._level:
            return True

# class _DomainBlock(NamedTuple):
#     domain: str # FIXME: Use an actual Domain object from somewhere?
#     severity: BlockSeverity = BlockSeverity.SUSPEND
#     public_comment: str = ''
#     private_comment: str = ''
#     reject_media: bool = False
#     reject_reports: bool = False
#     obfuscate: bool = False

class DomainBlock(object):

    fields = [
        'domain',
        'severity',
        'public_comment',
        'private_comment',
        'reject_media',
        'reject_reports',
        'obfuscate',
    ]

    all_fields = [
        'domain',
        'severity',
        'public_comment',
        'private_comment',
        'reject_media',
        'reject_reports',
        'obfuscate',
        'id'
    ]

    def __init__(self, domain:str,
            severity: BlockSeverity=BlockSeverity('suspend'),
            public_comment: str="",
            private_comment: str="",
            reject_media: bool=False,
            reject_reports: bool=False,
            obfuscate: bool=False,
            id: int=None):
        """Initialize the DomainBlock
        """        
        self.domain = domain
        self.severity = severity
        self.public_comment = public_comment
        self.private_comment = private_comment
        self.reject_media = reject_media
        self.reject_reports = reject_reports
        self.obfuscate = obfuscate
        self.id = id

    @property
    def severity(self):
        return self._severity

    @severity.setter
    def severity(self, sev):
        if isinstance(sev, BlockSeverity):
            self._severity = sev
        else:
            self._severity = BlockSeverity(sev)

    def _asdict(self):
        """Return a dict version of this object
        """
        dictval = {
            'domain': self.domain,
            'severity': str(self.severity),
            'public_comment': self.public_comment,
            'private_comment': self.private_comment,
            'reject_media': self.reject_media,
            'reject_reports': self.reject_reports,
            'obfuscate': self.obfuscate,
        }
        if self.id:
            dictval['id'] = self.id

        return dictval

    def compare_fields(self, other, fields=None)->list:
        """Compare two DomainBlocks on specific fields.
        If all the fields are equal, the DomainBlocks are equal.

        @returns: a list of the fields that are different
        """
        if not isinstance(other, DomainBlock):
            raise ValueError(f"Cannot compare DomainBlock to {type(other)}:{other}")

        if fields is None:
            fields = self.fields

        diffs = []
        # Check if all the fields are equal
        for field in self.fields:
            a = getattr(self, field)
            b = getattr(other, field)
            # log.debug(f"Comparing field {field}: '{a}' <> '{b}'")
            if getattr(self, field) != getattr(other, field):
                diffs.append(field)
        return diffs

    def __eq__(self, other):
        diffs = self.compare_fields(other)
        if len(diffs) == 0:
            return True

    def __repr__(self):

        return f"<DomainBlock {self._asdict()}>"

    def copy(self):
        """Make a copy of this object and return it
        """
        retval = DomainBlock(**self._asdict())
        return retval

    def update(self, dict):
        """Update my kwargs
        """
        for key in dict:
            setattr(self, key, dict[key])

    def __iter__(self):
        """Be iterable"""
        keys = self.fields

        if getattr(self, 'id', False):
            keys.append('id')

        for k in keys:
            yield k

    def __getitem__(self, k, default=None):
        "Behave like a dict for getting values"
        if k not in self.all_fields:
            raise KeyError(f"Invalid key '{k}'")

        return getattr(self, k, default)

    def get(self, k, default=None):
        return self.__getitem__(k, default)


from fediblockhole.const import BlockSeverity, SeverityLevel

def test_severity_eq():

    s1 = BlockSeverity('suspend')
    s2 = BlockSeverity('suspend')

    assert s1 == s2

    s3 = BlockSeverity('silence')
    s4 = BlockSeverity('silence')

    assert s3 == s4

    s5 = BlockSeverity('noop')
    s6 = BlockSeverity('noop')

    assert s5 == s6

def test_severity_ne():
    s1 = BlockSeverity('noop')
    s2 = BlockSeverity('silence')
    s3 = BlockSeverity('suspend')

    assert s1 != s2
    assert s2 != s3
    assert s1 != s3

def test_severity_lt():
    s1 = BlockSeverity('noop')
    s2 = BlockSeverity('silence')
    s3 = BlockSeverity('suspend')

    assert s1 < s2
    assert s2 < s3
    assert s1 < s3

def test_severity_gt():
    s1 = BlockSeverity('noop')
    s2 = BlockSeverity('silence')
    s3 = BlockSeverity('suspend')

    assert s2 > s1
    assert s3 > s2
    assert s3 > s1

def test_severity_le():
    s1 = BlockSeverity('noop')
    s2 = BlockSeverity('silence')
    s2a = BlockSeverity('silence')
    s3 = BlockSeverity('suspend')

    assert s1 <= s2
    assert s2a <= s2
    assert s2 <= s3
    assert s1 <= s3

def test_severity_ge():
    s1 = BlockSeverity('noop')
    s2 = BlockSeverity('silence')
    s2a = BlockSeverity('silence')
    s3 = BlockSeverity('suspend')

    assert s2 >= s1
    assert s2a >= s1
    assert s3 >= s2
    assert s3 >= s1




"""Test the config file is loading parameters correctly
"""
from util import shim_argparse
from fediblockhole import setup_argparse, augment_args

def test_parse_tomldata():
    tomldata = """
# Test TOML config for FediBlockHole

blocklist_instance_sources = []  

blocklist_url_sources = []  

save_intermediate = true

import_fields = ['public_comment']
"""
    ap = setup_argparse()
    args = ap.parse_args([])
    args = augment_args(args, tomldata)

    assert args.blocklist_instance_sources == []
    assert args.blocklist_url_sources == []
    assert args.save_intermediate == True
    assert args.import_fields == ['public_comment']

def test_set_mergeplan_max():
    tomldata = """mergeplan = 'max'
    """
    args = shim_argparse([], tomldata)

    assert args.mergeplan == 'max'

def test_set_mergeplan_min():
    tomldata = """mergeplan = 'min'
    """
    args = shim_argparse([], tomldata)

    assert args.mergeplan == 'min'

def test_set_allowlists():
    tomldata = """# Comment on config
allowlist_url_sources = [ { url='file:///path/to/allowlist', format='csv'} ] 
"""
    args = shim_argparse([], tomldata)

    assert args.mergeplan == 'max'
    assert args.allowlist_url_sources == [{
        'url': 'file:///path/to/allowlist',
        'format': 'csv',
        }]



"""Test the DomainBlock structure
"""
import pytest

from fediblockhole.const import DomainBlock, BlockSeverity, SeverityLevel

def test_blocksev_blankstring():
    a = BlockSeverity('')
    assert a.level == SeverityLevel.NONE

def test_blocksev_string_noop():
    a = BlockSeverity('noop')
    assert a.level == SeverityLevel.NONE

def test_blocksev_none():
    a = BlockSeverity(None)
    assert a.level == SeverityLevel.NONE

def test_empty_domainblock_fails():
    with pytest.raises(TypeError):
        a = DomainBlock()

def test_default_suspend():
    a = DomainBlock('example.org')
    assert a.domain == 'example.org'
    assert a.severity.level == SeverityLevel.SUSPEND

def test_severity_suspend():
    a = DomainBlock('example.org', 'suspend')
    assert a.domain == 'example.org'
    assert a.severity.level == SeverityLevel.SUSPEND

def test_severity_silence():
    a = DomainBlock('example.org', 'silence')
    assert a.domain == 'example.org'
    assert a.severity.level == SeverityLevel.SILENCE

def test_severity_noop_string():
    a = DomainBlock('example.org', 'noop')
    assert a.domain == 'example.org'
    assert a.severity.level == SeverityLevel.NONE

def test_severity_none():
    a = DomainBlock('example.org', None)
    assert a.domain == 'example.org'
    assert a.severity.level == SeverityLevel.NONE

def test_compare_equal_blocks():

    a = DomainBlock('example1.org', 'suspend')
    b = DomainBlock('example1.org', 'suspend')

    assert a == b

def test_compare_diff_domains():

    a = DomainBlock('example1.org', 'suspend')
    b = DomainBlock('example2.org', 'suspend')

    assert a != b

def test_compare_diff_sevs():

    a = DomainBlock('example1.org', 'suspend')
    b = DomainBlock('example1.org', 'silence')

    assert a != b

def test_compare_diff_sevs_2():

    a = DomainBlock('example1.org', 'suspend')
    b = DomainBlock('example1.org', 'noop')

    assert a != b



""" Test allowlists
"""
import pytest

from util import shim_argparse
from fediblockhole.const import DomainBlock
from fediblockhole import fetch_allowlists, apply_allowlists

def test_cmdline_allow_removes_domain():
    """Test that -A <domain> removes entries from merged
    """
    conf = shim_argparse(['-A', 'removeme.org'])

    merged = {
        'example.org': DomainBlock('example.org'),
        'example2.org': DomainBlock('example2.org'),
        'removeme.org': DomainBlock('removeme.org'),
        'keepblockingme.org': DomainBlock('keepblockingme.org'),
    }

    # allowlists = {
    #     'testlist': [ DomainBlock('removeme.org', 'noop'), ]
    # }
    
    merged = apply_allowlists(merged, conf, {})

    with pytest.raises(KeyError):
        merged['removeme.org']

def test_allowlist_removes_domain():
    """Test that an item in an allowlist removes entries from merged
    """
    conf = shim_argparse()

    merged = {
        'example.org': DomainBlock('example.org'),
        'example2.org': DomainBlock('example2.org'),
        'removeme.org': DomainBlock('removeme.org'),
        'keepblockingme.org': DomainBlock('keepblockingme.org'),
    }

    allowlists = {
        'testlist': [ DomainBlock('removeme.org', 'noop'), ]
    }
    
    merged = apply_allowlists(merged, conf, allowlists)

    with pytest.raises(KeyError):
        merged['removeme.org']

def test_allowlist_removes_tld():
    """Test that an item in an allowlist removes entries from merged
    """
    conf = shim_argparse()

    merged = {
        '.cf': DomainBlock('.cf'),
        'example.org': DomainBlock('example.org'),
        '.tk': DomainBlock('.tk'),
        'keepblockingme.org': DomainBlock('keepblockingme.org'),
    }

    allowlists = {
        'list1': [
            DomainBlock('.cf', 'noop'), 
            DomainBlock('.tk', 'noop'), 
        ]
    }
    
    merged = apply_allowlists(merged, conf, allowlists)

    with pytest.raises(KeyError):
        merged['.cf']

    with pytest.raises(KeyError):
        merged['.tk']


"""Test the commandline defined parameters correctly
"""
from util import shim_argparse
from fediblockhole import setup_argparse, augment_args

def test_cmdline_no_configfile():
    """ Test bare command with no configfile
    """
    ap = setup_argparse()
    args = ap.parse_args([])

    assert args.config == '/etc/default/fediblockhole.conf.toml'
    assert args.mergeplan == None
    assert args.blocklist_savefile == None
    assert args.save_intermediate == False
    assert args.savedir == None
    assert args.import_fields == None
    assert args.export_fields == None

    assert args.no_fetch_url == False
    assert args.no_fetch_instance == False
    assert args.no_push_instance == False
    assert args.dryrun == False

    assert args.loglevel == None

def test_cmdline_mergeplan_min():
    """ Test setting mergeplan min
    """
    ap = setup_argparse()
    args = ap.parse_args(['-m', 'min'])

    assert args.mergeplan == 'min'

def test_set_allow_domain():
    """Set a single allow domain on commandline"""
    ap = setup_argparse()
    args = ap.parse_args(['-A', 'example.org'])

    assert args.allow_domains == ['example.org']

def test_set_multiple_allow_domains():
    """Set multiple allow domains on commandline"""
    ap = setup_argparse()
    args = ap.parse_args(['-A', 'example.org', '-A', 'example2.org', '-A', 'example3.org'])

    assert args.allow_domains == ['example.org', 'example2.org', 'example3.org']


"""A tool for managing federated Mastodon blocklists
"""

import argparse
import toml
import csv
import requests
import json
import time
import os.path
import sys
import urllib.request as urlr

from .blocklist_parser import parse_blocklist
from .const import DomainBlock, BlockSeverity

from importlib.metadata import version
__version__ = version('fediblockhole')

import logging
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(message)s')
log = logging.getLogger('fediblockhole')

# Max size of a URL-fetched blocklist
URL_BLOCKLIST_MAXSIZE = 1024 ** 3

# Wait at most this long for a remote server to respond
REQUEST_TIMEOUT = 30

# Time to wait between instance API calls to we don't melt them
# The default Mastodon rate limit is 300 calls per 5 minutes
API_CALL_DELAY = 5 * 60 / 300 # 300 calls per 5 minutes

# We always import the domain and the severity
IMPORT_FIELDS = ['domain', 'severity']

# Allowlists always import these fields
ALLOWLIST_IMPORT_FIELDS = ['domain', 'severity', 'public_comment', 'private_comment', 'reject_media', 'reject_reports', 'obfuscate']

# We always export the domain and the severity
EXPORT_FIELDS = ['domain', 'severity']

def sync_blocklists(conf: argparse.Namespace):
    """Sync instance blocklists from remote sources.

    @param conf: A configuration dictionary
    """
    # Build a dict of blocklists we retrieve from remote sources.
    # We will merge these later using a merge algorithm we choose.

    # Always import these fields
    import_fields = IMPORT_FIELDS
    # Add extra import fields if defined in config
    import_fields.extend(conf.import_fields)

    # Always export these fields
    export_fields = EXPORT_FIELDS
    # Add extra export fields if defined in config
    export_fields.extend(conf.export_fields)

    blocklists = {}
    # Fetch blocklists from URLs
    if not conf.no_fetch_url:
        blocklists = fetch_from_urls(blocklists, conf.blocklist_url_sources,
            import_fields, conf.save_intermediate, conf.savedir, export_fields)

    # Fetch blocklists from remote instances
    if not conf.no_fetch_instance:
        blocklists = fetch_from_instances(blocklists, conf.blocklist_instance_sources,
            import_fields, conf.save_intermediate, conf.savedir, export_fields)

    # Merge blocklists into an update dict
    merged = merge_blocklists(blocklists, conf.mergeplan)

    # Remove items listed in allowlists, if any
    allowlists = fetch_allowlists(conf)
    merged = apply_allowlists(merged, conf, allowlists)

    # Save the final mergelist, if requested
    if conf.blocklist_savefile:
        log.info(f"Saving merged blocklist to {conf.blocklist_savefile}")
        save_blocklist_to_file(merged.values(), conf.blocklist_savefile, export_fields)

    # Push the blocklist to destination instances
    if not conf.no_push_instance:
        log.info("Pushing domain blocks to instances...")
        for dest in conf.blocklist_instance_destinations:
            domain = dest['domain']
            token = dest['token']
            max_followed_severity = BlockSeverity(dest.get('max_followed_severity', 'silence'))
            push_blocklist(token, domain, merged.values(), conf.dryrun, import_fields, max_followed_severity)

def apply_allowlists(merged: dict, conf: argparse.Namespace, allowlists: dict):
    """Apply allowlists
    """
    # Apply allows specified on the commandline
    for domain in conf.allow_domains:
        log.info(f"'{domain}' allowed by commandline, removing any blocks...")
        if domain in merged:
            del merged[domain]

    # Apply allows from URLs lists
    log.info("Removing domains from URL allowlists...")
    for key, alist in allowlists.items():
        log.debug(f"Processing allows from '{key}'...")
        for allowed in alist:
            domain = allowed.domain
            log.debug(f"Removing allowlisted domain '{domain}' from merged list.")
            if domain in merged:
                del merged[domain]

    return merged

def fetch_allowlists(conf: argparse.Namespace) -> dict:
    """
    """
    if conf.allowlist_url_sources:
        allowlists = fetch_from_urls({}, conf.allowlist_url_sources, ALLOWLIST_IMPORT_FIELDS)
        return allowlists
    return {}

def fetch_from_urls(blocklists: dict, url_sources: dict,
    import_fields: list=IMPORT_FIELDS,
    save_intermediate: bool=False,
    savedir: str=None, export_fields: list=EXPORT_FIELDS) -> dict:
    """Fetch blocklists from URL sources
    @param blocklists: A dict of existing blocklists, keyed by source
    @param url_sources: A dict of configuration info for url sources
    @returns: A dict of blocklists, same as input, but (possibly) modified
    """
    log.info("Fetching domain blocks from URLs...")

    for item in url_sources:
        url = item['url']
        # If import fields are provided, they override the global ones passed in
        source_import_fields = item.get('import_fields', None)
        if source_import_fields:
            # Ensure we always use the default fields
            import_fields = IMPORT_FIELDS.extend(source_import_fields)

        max_severity = item.get('max_severity', 'suspend')
        listformat = item.get('format', 'csv')
        with urlr.urlopen(url) as fp:
            rawdata = fp.read(URL_BLOCKLIST_MAXSIZE).decode('utf-8')
            blocklists[url] = parse_blocklist(rawdata, listformat, import_fields, max_severity)
            
        if save_intermediate:
            save_intermediate_blocklist(blocklists[url], url, savedir, export_fields)
    
    return blocklists

def fetch_from_instances(blocklists: dict, sources: dict,
    import_fields: list=IMPORT_FIELDS,
    save_intermediate: bool=False,
    savedir: str=None, export_fields: list=EXPORT_FIELDS) -> dict:
    """Fetch blocklists from other instances
    @param blocklists: A dict of existing blocklists, keyed by source
    @param url_sources: A dict of configuration info for url sources
    @returns: A dict of blocklists, same as input, but (possibly) modified
    """
    log.info("Fetching domain blocks from instances...")
    for item in sources:
        domain = item['domain']
        admin = item.get('admin', False)
        token = item.get('token', None)
        itemsrc = f"https://{domain}/api"

        # If import fields are provided, they override the global ones passed in
        source_import_fields = item.get('import_fields', None)
        if source_import_fields:
            # Ensure we always use the default fields
            import_fields = IMPORT_FIELDS.extend(source_import_fields)

        # Add the blocklist with the domain as the source key
        blocklists[itemsrc] = fetch_instance_blocklist(domain, token, admin, import_fields)
        if save_intermediate:
            save_intermediate_blocklist(blocklists[itemsrc], domain, savedir, export_fields)
    return blocklists

def merge_blocklists(blocklists: dict, mergeplan: str='max') -> dict:
    """Merge fetched remote blocklists into a bulk update
    @param blocklists: A dict of lists of DomainBlocks, keyed by source.
        Each value is a list of DomainBlocks
    @param mergeplan: An optional method of merging overlapping block definitions
        'max' (the default) uses the highest severity block found
        'min' uses the lowest severity block found
    @param returns: A dict of DomainBlocks keyed by domain
    """
    merged = {}

    for key, blist in blocklists.items():
        log.debug(f"processing blocklist from: {key} ...")
        for newblock in blist:
            domain = newblock.domain
            # If the domain has two asterisks in it, it's obfuscated
            # and we can't really use it, so skip it and do the next one
            if '*' in domain:
                log.debug(f"Domain '{domain}' is obfuscated. Skipping it.")
                continue

            elif domain in merged:
                log.debug(f"Overlapping block for domain {domain}. Merging...")
                blockdata = apply_mergeplan(merged[domain], newblock, mergeplan)

            else:
                # New block
                blockdata = newblock

            # end if
            log.debug(f"blockdata is: {blockdata}")
            merged[domain] = blockdata
        # end for
    return merged

def apply_mergeplan(oldblock: DomainBlock, newblock: DomainBlock, mergeplan: str='max') -> dict:
    """Use a mergeplan to decide how to merge two overlapping block definitions
    
    @param oldblock: The existing block definition.
    @param newblock: The new block definition we want to merge in.
    @param mergeplan: How to merge. Choices are 'max', the default, and 'min'.
    """
    # Default to the existing block definition
    blockdata = oldblock._asdict()

    # Merge comments
    keylist = ['public_comment', 'private_comment']
    for key in keylist:
        try:
            oldcomment = getattr(oldblock, key)
            newcomment = getattr(newblock, key)
            blockdata[key] = merge_comments(oldcomment, newcomment)
        except KeyError:
            log.debug(f"Key '{key}' missing from block definition so cannot compare. Continuing...")
            continue
    
    # How do we override an earlier block definition?
    if mergeplan in ['max', None]:
        # Use the highest block level found (the default)
        log.debug(f"Using 'max' mergeplan.")

        if newblock.severity > oldblock.severity:
            log.debug(f"New block severity is higher. Using that.")
            blockdata['severity'] = newblock.severity
        
        # For 'reject_media', 'reject_reports', and 'obfuscate' if
        # the value is set and is True for the domain in
        # any blocklist then the value is set to True.
        for key in ['reject_media', 'reject_reports', 'obfuscate']:
            newval = getattr(newblock, key)
            if newval == True:
                blockdata[key] = True

    elif mergeplan in ['min']:
        # Use the lowest block level found
        log.debug(f"Using 'min' mergeplan.")

        if newblock.severity < oldblock.severity:
            blockdata['severity'] = newblock.severity

        # For 'reject_media', 'reject_reports', and 'obfuscate' if
        # the value is set and is False for the domain in
        # any blocklist then the value is set to False.
        for key in ['reject_media', 'reject_reports', 'obfuscate']:
            newval = getattr(newblock, key)
            if newval == False:
                blockdata[key] = False

    else:
        raise NotImplementedError(f"Mergeplan '{mergeplan}' not implemented.")

    log.debug(f"Block severity set to {blockdata['severity']}")

    return DomainBlock(**blockdata)

def merge_comments(oldcomment:str, newcomment:str) -> str:
    """ Merge two comments

    @param oldcomment: The original comment we're merging into
    @param newcomment: The new commment we want to merge in
    @returns: a new str of the merged comment
    """
    # Don't merge if both comments are None or ''
    if oldcomment in ['', None] and newcomment in ['', None]:
        return ''

    # If both comments are the same, don't merge
    if oldcomment == newcomment:
        return oldcomment

    # We want to skip duplicate fragments so we don't end up
    # re-concatenating the same strings every time there's an
    # update, causing the comment to grow without bound.
    # We tokenize the comments, splitting them on ', ', and comparing
    # the tokens, skipping duplicates.
    # This means "boring, lack of moderation, nazis, scrapers" merging
    # with "lack of moderation, scrapers" should result in
    # "boring, lack of moderation, nazis, scrapers"
    old_tokens = oldcomment.split(', ')
    new_tokens = newcomment.split(', ')
    
    # Remove any empty string tokens that we get
    while '' in old_tokens:
        old_tokens.remove('')
    while '' in new_tokens:
        new_tokens.remove('')

    # Remove duplicate tokens
    for token in old_tokens:
        if token in new_tokens:
            new_tokens.remove(token)

    # Combine whatever tokens are left into one set
    tokenset = old_tokens
    tokenset.extend(new_tokens)

    # Return the merged string
    return ', '.join(tokenset)

def requests_headers(token: str=None):
    """Set common headers for requests"""
    headers = {
        'User-Agent': f"FediBlockHole/{__version__}"
    }
    if token:
        headers['Authorization'] = f"Bearer {token}"

    return headers

def fetch_instance_blocklist(host: str, token: str=None, admin: bool=False,
    import_fields: list=['domain', 'severity']) -> list[DomainBlock]:
    """Fetch existing block list from server

    @param host: The remote host to connect to.
    @param token: The (optional) OAuth Bearer token to authenticate with.
    @param admin: Boolean flag to use the admin API if True.
    @param import_fields: A list of fields to import from the remote instance.
    @returns: A list of the domain blocks from the instance.
    """
    log.info(f"Fetching instance blocklist from {host} ...")

    if admin:
        api_path = "/api/v1/admin/domain_blocks"
    else:
        api_path = "/api/v1/instance/domain_blocks"

    headers = requests_headers(token)

    url = f"https://{host}{api_path}"

    blocklist = []
    link = True

    while link:
        response = requests.get(url, headers=headers, timeout=REQUEST_TIMEOUT)
        if response.status_code != 200:
            log.error(f"Cannot fetch remote blocklist: {response.content}")
            raise ValueError("Unable to fetch domain block list: %s", response)

        blocklist.extend( parse_blocklist(response.content, 'json', import_fields) )
        
        # Parse the link header to find the next url to fetch
        # This is a weird and janky way of doing pagination but
        # hey nothing we can do about it we just have to deal
        link = response.headers.get('Link', None)
        if link is None:
            break
        pagination = link.split(', ')
        if len(pagination) != 2:
            link = None
            break
        else:
            next = pagination[0]
            # prev = pagination[1]
        
            urlstring, rel = next.split('; ')
            url = urlstring.strip('<').rstrip('>')

    return blocklist

def delete_block(token: str, host: str, id: int):
    """Remove a domain block"""
    log.debug(f"Removing domain block {id} at {host}...")
    api_path = "/api/v1/admin/domain_blocks/"

    url = f"https://{host}{api_path}{id}"

    response = requests.delete(url,
        headers=requests_headers(token),
        timeout=REQUEST_TIMEOUT
    )
    if response.status_code != 200:
        if response.status_code == 404:
            log.warning(f"No such domain block: {id}")
            return

        raise ValueError(f"Something went wrong: {response.status_code}: {response.content}")

def fetch_instance_follows(token: str, host: str, domain: str) -> int:
    """Fetch the followers of the target domain at the instance

    @param token: the Bearer authentication token for OAuth access
    @param host: the instance API hostname/IP address
    @param domain: the domain to search for followers of
    @returns: int, number of local followers of remote instance accounts
    """
    api_path = "/api/v1/admin/measures"
    url = f"https://{host}{api_path}"

    key = 'instance_follows'

    # This data structure only allows us to request a single domain
    # at a time, which limits the load on the remote instance of each call
    data = {
        'keys': [
            key
            ],
        key: { 'domain': domain },
    }

    # The Mastodon API only accepts JSON formatted POST data for measures
    response = requests.post(url,
        headers=requests_headers(token),
        json=data,
        timeout=REQUEST_TIMEOUT
    )
    if response.status_code != 200:
        if response.status_code == 403:
            log.error(f"Cannot fetch follow information for {domain} from {host}: {response.content}")

        raise ValueError(f"Something went wrong: {response.status_code}: {response.content}")

    # Get the total returned
    follows = int(response.json()[0]['total'])
    return follows

def check_followed_severity(host: str, token: str, domain: str,
    severity: BlockSeverity,
    max_followed_severity: BlockSeverity=BlockSeverity('silence')):
    """Check an instance to see if it has followers of a to-be-blocked instance"""

    log.debug("Checking followed severity...")
    # Return straight away if we're not increasing the severity
    if severity <= max_followed_severity:
        return severity

    # If the instance has accounts that follow people on the to-be-blocked domain,
    # limit the maximum severity to the configured `max_followed_severity`.
    log.debug("checking for instance follows...")
    follows = fetch_instance_follows(token, host, domain)
    time.sleep(API_CALL_DELAY)
    if follows > 0:
        log.debug(f"Instance {host} has {follows} followers of accounts at {domain}.")
        if severity > max_followed_severity:
            log.warning(f"Instance {host} has {follows} followers of accounts at {domain}. Limiting block severity to {max_followed_severity}.")
            return max_followed_severity
    return severity

def is_change_needed(oldblock: dict, newblock: dict, import_fields: list):
    change_needed = oldblock.compare_fields(newblock, import_fields)
    return change_needed

def update_known_block(token: str, host: str, block: DomainBlock):
    """Update an existing domain block with information in blockdict"""
    api_path = "/api/v1/admin/domain_blocks/"

    try:
        id = block.id
        blockdata = block._asdict()
        del blockdata['id']
    except KeyError:
        import pdb
        pdb.set_trace()

    url = f"https://{host}{api_path}{id}"

    response = requests.put(url,
        headers=requests_headers(token),
        json=blockdata._asdict(),
        timeout=REQUEST_TIMEOUT
    )
    if response.status_code != 200:
        raise ValueError(f"Something went wrong: {response.status_code}: {response.content}")

def add_block(token: str, host: str, blockdata: DomainBlock):
    """Block a domain on Mastodon host
    """
    log.debug(f"Adding block entry for {blockdata.domain} at {host}...")
    api_path = "/api/v1/admin/domain_blocks"

    url = f"https://{host}{api_path}"

    response = requests.post(url,
        headers=requests_headers(token),
        json=blockdata._asdict(),
        timeout=REQUEST_TIMEOUT
    )
    if response.status_code == 422:
        # A stricter block already exists. Probably for the base domain.
        err = json.loads(response.content)
        log.warning(err['error'])

    elif response.status_code != 200:
            
        raise ValueError(f"Something went wrong: {response.status_code}: {response.content}")
           
def push_blocklist(token: str, host: str, blocklist: list[dict],
                    dryrun: bool=False,
                    import_fields: list=['domain', 'severity'],
                    max_followed_severity:BlockSeverity=BlockSeverity('silence'),
                    ):
    """Push a blocklist to a remote instance.
    
    Merging the blocklist with the existing list the instance has,
    updating existing entries if they exist.

    @param token: The Bearer token for OAUTH API authentication
    @param host: The instance host, FQDN or IP
    @param blocklist: A list of block definitions. They must include the domain.
    @param import_fields: A list of fields to import to the instances.
    """
    log.info(f"Pushing blocklist to host {host} ...")
    # Fetch the existing blocklist from the instance
    # Force use of the admin API, and add 'id' to the list of fields
    if 'id' not in import_fields:
        import_fields.append('id')
    serverblocks = fetch_instance_blocklist(host, token, True, import_fields)

    # # Convert serverblocks to a dictionary keyed by domain name
    knownblocks = {row.domain: row for row in serverblocks}

    for newblock in blocklist:

        log.debug(f"Processing block: {newblock}")
        oldblock = knownblocks.get(newblock.domain, None)
        if oldblock:
            log.debug(f"Block already exists for {newblock.domain}, checking for differences...")

            change_needed = is_change_needed(oldblock, newblock, import_fields)

            # Is the severity changing?
            if 'severity' in change_needed:
                log.debug("Severity change requested, checking...")
                if newblock.severity > oldblock.severity:
                    # Confirm if we really want to change the severity
                    # If we still have followers of the remote domain, we may not
                    # want to go all the way to full suspend, depending on the configuration
                    newseverity = check_followed_severity(host, token, oldblock.domain, newblock.severity, max_followed_severity)
                    if newseverity != oldblock.severity:
                        newblock.severity = newseverity
                    else:
                        log.info("Keeping severity of block the same to avoid disrupting followers.")
                        change_needed.remove('severity')

            if change_needed:
                log.info(f"Change detected. Need to update {change_needed} for domain block for {oldblock.domain}")
                log.info(f"Old block definition: {oldblock}")
                log.info(f"Pushing new block definition: {newblock}")
                blockdata = oldblock.copy()
                blockdata.update(newblock)
                log.debug(f"Block as dict: {blockdata._asdict()}")

                if not dryrun:
                    update_known_block(token, host, blockdata)
                    # add a pause here so we don't melt the instance
                    time.sleep(API_CALL_DELAY)
                else:
                    log.info("Dry run selected. Not applying changes.")

            else:
                log.debug("No differences detected. Not updating.")
                pass

        else:
            # This is a new block for the target instance, so we
            # need to add a block rather than update an existing one
            log.info(f"Adding new block: {newblock}...")
            log.debug(f"Block as dict: {newblock._asdict()}")

            # Make sure the new block doesn't clobber a domain with followers
            newblock.severity = check_followed_severity(host, token, newblock.domain, newblock.severity, max_followed_severity)
            if not dryrun:
                add_block(token, host, newblock)
                # add a pause here so we don't melt the instance
                time.sleep(API_CALL_DELAY)
            else:
                log.info("Dry run selected. Not adding block.")

def load_config(configfile: str):
    """Augment commandline arguments with config file parameters
    
    Config file is expected to be in TOML format
    """
    conf = toml.load(configfile)
    return conf

def save_intermediate_blocklist(
    blocklist: list[dict], source: str,
    filedir: str,
    export_fields: list=['domain','severity']):
    """Save a local copy of a blocklist we've downloaded
    """
    # Invent a filename based on the remote source
    # If the source was a URL, convert it to something less messy
    # If the source was a remote domain, just use the name of the domain
    log.debug(f"Saving intermediate blocklist from {source}")
    source = source.replace('/','-')
    filename = f"{source}.csv"
    filepath = os.path.join(filedir, filename)
    save_blocklist_to_file(blocklist, filepath, export_fields)

def save_blocklist_to_file(
    blocklist: list[DomainBlock],
    filepath: str,
    export_fields: list=['domain','severity']):
    """Save a blocklist we've downloaded from a remote source

    @param blocklist: A dictionary of block definitions, keyed by domain
    @param filepath: The path to the file the list should be saved in.
    @param export_fields: Which fields to include in the export.
    """
    try:
        blocklist = sorted(blocklist, key=lambda x: x.domain)
    except KeyError:
        log.error("Field 'domain' not found in blocklist.")
        log.debug(f"blocklist is: {blocklist}")

    log.debug(f"export fields: {export_fields}")

    with open(filepath, "w") as fp:
        writer = csv.DictWriter(fp, export_fields, extrasaction='ignore')
        writer.writeheader()
        for item in blocklist:
            writer.writerow(item._asdict())

def augment_args(args, tomldata: str=None):
    """Augment commandline arguments with config file parameters
    
    If tomldata is provided, uses that data instead of loading
    from a config file.
    """
    if tomldata:
        conf = toml.loads(tomldata)
    else:
        conf = toml.load(args.config)

    if not args.no_fetch_url:
        args.no_fetch_url = conf.get('no_fetch_url', False)

    if not args.no_fetch_instance:
        args.no_fetch_instance = conf.get('no_fetch_instance', False)

    if not args.no_push_instance:
        args.no_push_instance = conf.get('no_push_instance', False)

    if not args.blocklist_savefile:
        args.blocklist_savefile = conf.get('blocklist_savefile', None)

    if not args.save_intermediate:
        args.save_intermediate = conf.get('save_intermediate', False)
    
    if not args.savedir:
        args.savedir = conf.get('savedir', '/tmp')

    if not args.export_fields:
        args.export_fields = conf.get('export_fields', [])

    if not args.import_fields:
        args.import_fields = conf.get('import_fields', [])

    if not args.mergeplan:
        args.mergeplan = conf.get('mergeplan', 'max')

    args.blocklist_url_sources = conf.get('blocklist_url_sources', [])
    args.blocklist_instance_sources = conf.get('blocklist_instance_sources', [])
    args.allowlist_url_sources = conf.get('allowlist_url_sources', [])
    args.blocklist_instance_destinations = conf.get('blocklist_instance_destinations', [])

    return args

def setup_argparse():
    """Setup the commandline arguments
    """
    ap = argparse.ArgumentParser(
        description="Bulk blocklist tool",
        epilog=f"Part of FediBlockHole v{__version__}",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument('-c', '--config', default='/etc/default/fediblockhole.conf.toml', help="Config file")
    ap.add_argument('-V', '--version', action='store_true', help="Show version and exit.")

    ap.add_argument('-o', '--outfile', dest="blocklist_savefile", help="Save merged blocklist to a local file.")
    ap.add_argument('-S', '--save-intermediate', dest="save_intermediate", action='store_true', help="Save intermediate blocklists we fetch to local files.")
    ap.add_argument('-D', '--savedir', dest="savedir", help="Directory path to save intermediate lists.")
    ap.add_argument('-m', '--mergeplan', choices=['min', 'max'], help="Set mergeplan.")

    ap.add_argument('-I', '--import-field', dest='import_fields', action='append', help="Extra blocklist fields to import.")
    ap.add_argument('-E', '--export-field', dest='export_fields', action='append', help="Extra blocklist fields to export.")
    ap.add_argument('-A', '--allow', dest="allow_domains", action='append', default=[], help="Override any blocks to allow this domain.")

    ap.add_argument('--no-fetch-url', dest='no_fetch_url', action='store_true', help="Don't fetch from URLs, even if configured.")
    ap.add_argument('--no-fetch-instance', dest='no_fetch_instance', action='store_true', help="Don't fetch from instances, even if configured.")
    ap.add_argument('--no-push-instance', dest='no_push_instance', action='store_true', help="Don't push to instances, even if configured.")

    ap.add_argument('--loglevel', choices=['debug', 'info', 'warning', 'error', 'critical'], help="Set log output level.")
    ap.add_argument('--dryrun', action='store_true', help="Don't actually push updates, just show what would happen.")

    return ap

def main():

    ap = setup_argparse()
    args = ap.parse_args()

    if args.loglevel is not None:
        levelname = args.loglevel.upper()
        log.setLevel(getattr(logging, levelname))

    if args.version:
        print(f"v{__version__}")
        sys.exit(0)

    # Load the configuration file
    args = augment_args(args)

    # Do the work of syncing
    sync_blocklists(args)



"""Test parsing the RapidBlock JSON format
"""
from fediblockhole.blocklist_parser import parse_blocklist

from fediblockhole.const import SeverityLevel

rapidblockjson = "data-rapidblock.json"

def test_parse_rapidblock_json():
    with open(rapidblockjson) as fp:
        data = fp.read()
        bl = parse_blocklist(data, 'rapidblock.json')

        assert bl[0].domain == '101010.pl'
        assert bl[0].severity.level == SeverityLevel.SUSPEND
        assert bl[0].public_comment == ''

        assert bl[10].domain == 'berserker.town'
        assert bl[10].severity.level == SeverityLevel.SUSPEND
        assert bl[10].public_comment == ''
        assert bl[10].private_comment == ''

def test_parse_with_comments():
    with open(rapidblockjson) as fp:
        data = fp.read()
        bl = parse_blocklist(data, 'rapidblock.json', ['domain', 'severity', 'public_comment', 'private_comment'])

        assert bl[0].domain == '101010.pl'
        assert bl[0].severity.level == SeverityLevel.SUSPEND
        assert bl[0].public_comment == 'cryptomining javascript, white supremacy'

        assert bl[10].domain == 'berserker.town'
        assert bl[10].severity.level == SeverityLevel.SUSPEND
        assert bl[10].public_comment == 'freeze peach'


"""Tests of the CSV parsing
"""

from fediblockhole.blocklist_parser import BlocklistParserCSV, parse_blocklist
from fediblockhole.const import DomainBlock, BlockSeverity, SeverityLevel


def test_single_line():
    csvdata = "example.org"

    parser = BlocklistParserCSV()
    bl = parser.parse_blocklist(csvdata)
    assert len(bl) == 0

def test_header_only():
    csvdata = "domain,severity,public_comment"

    parser = BlocklistParserCSV()
    bl = parser.parse_blocklist(csvdata)
    assert len(bl) == 0

def test_2_blocks():
    csvdata = """domain,severity
example.org,silence
example2.org,suspend
"""

    parser = BlocklistParserCSV()
    bl = parser.parse_blocklist(csvdata)

    assert len(bl) == 2
    assert bl[0].domain == 'example.org'

def test_4_blocks():
    csvdata = """domain,severity,public_comment
example.org,silence,"test 1"
example2.org,suspend,"test 2"
example3.org,noop,"test 3"
example4.org,suspend,"test 4"
"""

    parser = BlocklistParserCSV()
    bl = parser.parse_blocklist(csvdata)

    assert len(bl) == 4
    assert bl[0].domain == 'example.org'
    assert bl[1].domain == 'example2.org'
    assert bl[2].domain == 'example3.org'
    assert bl[3].domain == 'example4.org'

    assert bl[0].severity.level == SeverityLevel.SILENCE
    assert bl[1].severity.level == SeverityLevel.SUSPEND
    assert bl[2].severity.level == SeverityLevel.NONE
    assert bl[3].severity.level == SeverityLevel.SUSPEND

def test_ignore_comments():
    csvdata = """domain,severity,public_comment,private_comment
example.org,silence,"test 1","ignore me"
example2.org,suspend,"test 2","ignote me also"
example3.org,noop,"test 3","and me"
example4.org,suspend,"test 4","also me"
"""

    parser = BlocklistParserCSV()
    bl = parser.parse_blocklist(csvdata)

    assert len(bl) == 4
    assert bl[0].domain == 'example.org'
    assert bl[1].domain == 'example2.org'
    assert bl[2].domain == 'example3.org'
    assert bl[3].domain == 'example4.org'

    assert bl[0].public_comment == ''
    assert bl[0].private_comment == ''

    assert bl[2].public_comment == ''
    assert bl[2].private_comment == ''